/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <stdio.h>
#include <string.h>

#include "elster_instance.h"

#include "NTypes.h"
#include "NUtils.h"

static KCanElster * Can = NULL;

bool CanIsInit = false;
bool trace = false;
enum KCanServer::dt_type drv_type = KCanServer::dt_can;
char can_dev[128] = "can0";
char err_msg[128];

bool elster_instance::InitCan()
{
  if (CanIsInit)
    return true;
  
  err_msg[0] = 0;
  if (!Can)
    Can = new KCanElster;

  Can->driver_type = drv_type;
  if (!Can->Init(can_dev, 0x680))
  {
    sprintf(err_msg, "device \"%s\" not connected\n", can_dev);

    if (trace)
      printf("%s", err_msg);
    return false;
  }
  CanIsInit = true;
  Can->Resume();

  if (trace)
    printf("can connected\n");

  return true;
}

void elster_instance::undef()
{
  if (Can)
  {
    Can->Halt();
    NUtils::SleepMs(1000);
    
    delete Can;
  }
  Can = NULL;
  CanIsInit = false;
}

bool elster_instance::elsterInit(unsigned short send_id)
{
  if (send_id != 0x680 && send_id != 0x700 && send_id != 0x780)
  {
    strcpy(err_msg, "not allowed send_id");
    return false;
  }

  if (!CanIsInit)
    if (!InitCan())
      return false;

  Can->Trace = trace;
  Can->SendCanId = send_id;

  return true;
}

unsigned short elster_instance::get_hex(const char * & params, bool & err)
{
  while (*params == ' ')
    params++;

  if (NUtils::IsHexDigit(*params))
  {
    return (unsigned short) NUtils::GetHex(params);
  } else
    err = true;

  return 0;
}

bool elster_instance::test_end(const char * & params, bool & err)
{
  if (err)
    strcpy(err_msg, "invalid parameter");
  else
    *err_msg = 0;

  while (*params == ' ')
    params++;

  if (!err && *params)
  {
    err = true;
    strcpy(err_msg, "too much parameters");
  }

  return err;
}

const char * elster_instance::GetOneParam(const char * params, unsigned short & value)
{
  bool err = false;
  value = get_hex(params, err);

  if (test_end(params, err))
    return GetErrMsg();

  return NULL;
}

const char * elster_instance::GetThreeParams(const char * params, unsigned short & value1,
                            unsigned short & value2, unsigned short & value3)
{
  bool err = false;
  value1 = get_hex(params, err);
  value2 = get_hex(params, err);
  value3 = get_hex(params, err);

  if (test_end(params, err))
    return GetErrMsg();

  return NULL;
}

const char * elster_instance::GetFourParams(const char * params, unsigned short & value1,
                           unsigned short & value2, unsigned short & value3,
                           unsigned short & value4)
{
  bool err = false;
  value1 = get_hex(params, err);
  value2 = get_hex(params, err);
  value3 = get_hex(params, err);
  value4 = get_hex(params, err);

  if (test_end(params, err))
    return GetErrMsg();

  return NULL;
}

const char * elster_instance::GetErrMsg()
{
  return err_msg;
}

const char * elster_instance::toggle_trace(void)
{
  static char result[32];

  trace = !trace;
  strcpy(result, trace ? "true" : "false");
  return result;
}

const char * elster_instance::setdev(const char * dev)
{
  if (strlen(dev) >= High(can_dev))
    return "";

  strcpy(can_dev, dev);
  return can_dev;
}

const char * elster_instance::setcs(void)
{
  if (CanIsInit)
    return "";

  drv_type = KCanServer::dt_cs;
  return "ok";
}

const char * elster_instance::set_can232(void)
{
  if (CanIsInit)
    return "";
  
  drv_type = KCanServer::dt_can232;
  return "ok";
}

bool elster_instance::GetValue(unsigned short receiver_id, unsigned short elster_idx, unsigned short & Value)
{
  if (!CanIsInit)
    return false;
  
  return Can->GetValue(receiver_id, elster_idx, Value);
}

bool elster_instance::SetValue(unsigned short receiver_id, unsigned short elster_idx, unsigned short Value)
{
  if (!CanIsInit)
    return false;
 
  return Can->SetValue(receiver_id, elster_idx, Value);
}

bool elster_instance::GetDoubleValue(unsigned short first_val, unsigned scan_can_id,
                                     unsigned short elster_idx, unsigned char elster_type, double & result)
{
  if (!CanIsInit)
    return false;
  
  return Can->GetDoubleValue(first_val, scan_can_id, elster_idx, elster_type, result);
}

