/*
 *
 *  Copyright (C) 2015 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

// Filtert aus einem log-File, das mit ComfortSoft und CS_Bruecke erstellt wurde,
// für die Wärmepumpe spezifischen Angaben heraus und speichert sie in fhem_tab.inc.
// Diese Angaben sollten im File 50_Elster.pm eingesetzt werden.

#include <string.h>

#include "VList.h"
#include "NUtils.h"
#include "KElsterTable.h"

int main(int argc, const char * argv[])
{
  VList list;
  const char * Name = "cs_log.txt";
  
  if (argc == 2)
    Name = argv[1];
  
  if (!list.ReadFile(Name))
  {
    printf("\nFile '%s' not found\n\n", Name);
    printf("Use: LogAnalize <filename>\n\n");
    return -1;
  }

  for (int i = list.string_count; i >= 0; i--)
  {
    const char * str = list.GetString(i);
    if (str && strstr(str, "Elster ") != str)
      list.DeleteString(i);
  }
  list.Sort();
  
  char new_name[1024];
  const char * file_name = NUtils::ExtractFilename(Name);
  
  new_name[0] = 0;
  if (Name != file_name)
  {
    strncpy(new_name, Name, strlen(Name) - strlen(file_name));
    new_name[strlen(Name) - strlen(file_name)] = 0;
  }
  strcat(new_name, "sorted_");
  strcat(new_name, file_name);
  
//  if (!list.SaveFile(new_name))
//    printf("'%s' not saved\n", new_name);
  
  KStream tab;
  bool first = true;
  for (int i = 0; i < list.string_count; i++)
  {
    const char * p = list.GetString(i);
    if (p)
    {
      char dev[16];
      char id[16];
      char longdev[32];
      char str[256];
      char temp[256];
      bool sub = false;
      
      if (strstr(p, "= not available"))
        continue;
      
      p += 7;
      strncpy(dev, p, 3);
      dev[3] = 0;
      strncpy(id, p+4, 4);
      id[4] = 0;
      if (dev[0] == '0')
        strcpy(longdev, "$direct");
      else
      if (!strcmp(dev, "180"))
        strcpy(longdev, "$kessel");
      else
      if (dev[0] == '3' && dev[1] == '0')
      {
        strcpy(longdev, "$bedien_");
        strcat(longdev, dev + 2);
        sub = true;
      } else
      if (!strcmp(dev, "480"))
        strcpy(longdev, "$manager");
      else
      if (!strcmp(dev, "500"))
        strcpy(longdev, "$heiz");
      else
      if (dev[0] == '6' && dev[1] == '0')
      {
        strcpy(longdev, "$mischer_");
        strcat(longdev, dev + 2);
        sub = true;
      } else
        strcpy(longdev, dev);
      
      strcpy(temp, p+15);
      int i = 0;
      while (temp[i] && temp[i] != ' ')
        i++;
      temp[i] = 0;
      if (sub)
      {
        sprintf(temp + strlen(temp), "_%c", dev[2]);
      }
      
      const char * p_id = id;
      unsigned short elster_index = NUtils::GetHex(p_id);
      const ElsterIndex * ei = GetElsterIndex(elster_index);
      const ElsterIndex * ei1 = GetElsterIndex(elster_index+1);
      if (ei && ei1)
      {
        if (ei->Type == et_default && ei1->Type == et_double_val)
          continue;
      
        if (ei->Type == et_double_val && ei1->Type == et_triple_val)
          continue;
      }
      
      if (!first)
      {
        tab.AppendString(",\n");  
      }
      sprintf(str, "  \"%s\"   => { EIdx => \"%s\", ID => \"%s\" } ", temp, id, dev);
      tab.AppendString(str);
      first = false;
    }
  }
  
  new_name[0] = 0;
  if (Name != file_name)
  {
    strncpy(new_name, Name, strlen(Name) - strlen(file_name));
    new_name[strlen(Name) - strlen(file_name)] = 0;
  }
  strcat(new_name, "fhem_tab.inc");
  if (tab.SaveFile(new_name))
    printf("\nFile '%s' saved\n\n", new_name);
  
  return 0;
}

